import pytest
try:
    from embiggen.utils.tensorflow_utils import (
        tensorflow_version_is_higher_or_equal_than,
        tensorflow_version_is_less_or_equal_than,
        must_have_tensorflow_version_higher_or_equal_than,
        has_single_gpu
    )
    from embiggen.utils.normalize_model_structural_parameters import normalize_model_list_parameter, normalize_model_ragged_list_parameter
    from unittest import TestCase


    class TestTensorFlowUtils(TestCase):

        def setUp(self):
            pass

        def test_normalize_model_list_parameter(self):
            values = normalize_model_list_parameter(
                10,
                3,
                int
            )

            self.assertEqual(values, [10, 10, 10])

            with pytest.raises(ValueError):
                normalize_model_list_parameter([], 0, int)

            with pytest.raises(ValueError):
                normalize_model_list_parameter(8.0, 5, int)

            with pytest.raises(ValueError):
                normalize_model_list_parameter([2], 5, int)

            with pytest.raises(ValueError):
                normalize_model_list_parameter([2.0], 1, int)

        def test_normalize_model_ragged_list_parameter(self):
            normalize_model_ragged_list_parameter(
                None,
                3,
                layers_per_submodule=[1, 2, 3],
                object_type=int,
                default_value=6
            )

        def test_tensorflow_version_is_higher_or_equal_than(self):
            try:
                with pytest.raises(ValueError):
                    tensorflow_version_is_higher_or_equal_than("not a version")
                with pytest.raises(ValueError):
                    tensorflow_version_is_less_or_equal_than("not a version")
                self.assertTrue(tensorflow_version_is_higher_or_equal_than("0.0.0"))
                self.assertTrue(tensorflow_version_is_less_or_equal_than("99999.9.9"))
                must_have_tensorflow_version_higher_or_equal_than("0.0.0")
                with pytest.raises(ValueError):
                    must_have_tensorflow_version_higher_or_equal_than("not a version")
                with pytest.raises(ValueError):
                    must_have_tensorflow_version_higher_or_equal_than("9999.99.9")
            except (ValueError):
                pass
            has_single_gpu()
except (ModuleNotFoundError):
    pass


"""Module providing Perceptron for edge prediction."""
import json
from typing import Any, Dict, List, Optional, Union

import compress_json
import compress_pickle
import numpy as np
from ensmallen import Graph, models

from embiggen.edge_prediction.edge_prediction_model import \
    AbstractEdgePredictionModel
from embiggen.embedding_transformers import NodeTransformer


class PerceptronEdgePrediction(AbstractEdgePredictionModel):
    """Perceptron model for edge prediction."""

    def __init__(
        self,
        edge_features: Optional[Union[str, List[str]]] = "JaccardCoefficient",
        edge_embeddings: Optional[Union[str, List[str]]] = None,
        cooccurrence_iterations: int = 100,
        cooccurrence_window_size: int = 10,
        number_of_epochs: int = 1000,
        number_of_edges_per_mini_batch: int = 4096,
        sample_only_edges_with_heterogeneous_node_types: bool = False,
        learning_rate: float = 0.001,
        first_order_decay_factor: float = 0.9,
        second_order_decay_factor: float = 0.999,
        avoid_false_negatives: bool = False,
        use_scale_free_distribution: bool = True,
        random_state: int = 42,
        verbose: bool = True
    ):
        """Create new Perceptron object.

        Parameters
        ------------------------
        edge_features: Optional[Union[str, List[str]]] = "JaccardCoefficient"
            The edge features to compute for each edge.
            Zero or more edge features can be used at once.
            The currently supported edge features are:
            - Degree,
            - AdamicAdar,
            - JaccardCoefficient,
            - Cooccurrence,
            - ResourceAllocationIndex,
            - PreferentialAttachment,
        edge_embeddings: Optional[Union[str, List[str]]] = None
            The embedding methods to use for the provided node features.
            Zero or more edge emmbedding methods can be used at once.
            The currently supported edge embedding are:
            - CosineSimilarity,
            - EuclideanDistance,
            - Concatenate,
            - Hadamard,
            - L1,
            - L2,
            - Add,
            - Sub,
            - Maximum,
            - Minimum,
        cooccurrence_iterations: int = 100
            Number of iterations to run when computing the cooccurrence metric.
            By default 100.
        cooccurrence_window_size: int = 10
            Window size to consider to measure the cooccurrence.
            By default 100.
        number_of_epochs: int = 1000
            The number of epochs to train the model for. By default, 100.
        number_of_edges_per_mini_batch: int = 4096
            The number of samples to include for each mini-batch. By default 4096.
        sample_only_edges_with_heterogeneous_node_types: bool = False
            Whether to sample negative edges only with source and
            destination nodes that have different node types. By default false.
        learning_rate: float = 0.001
            Learning rate to use while training the model.
            By default 0.001.
        first_order_decay_factor: float = 0.9
            First order decay factor for the first order momentum.
            By default 0.9.
        second_order_decay_factor: float = 0.999
            Second order decay factor for the second order momentum.
            By default 0.999.
        avoid_false_negatives: bool = False
            Whether to avoid sampling false negatives.
            This may cause a slower training.
        use_scale_free_distribution: bool = True
            Whether to train model using a scale free distribution for the negatives.
        random_state: int = 42
            The random state to reproduce the model initialization and training. By default, 42.
        verbose: bool = True
            Whether to show epochs loading bar.
        """
        super().__init__(random_state=random_state)

        if isinstance(edge_features, str):
            edge_features = [edge_features]

        if isinstance(edge_embeddings, str):
            edge_embeddings = [edge_embeddings]

        self._model_kwargs = dict(
            edge_features=edge_features,
            edge_embeddings=edge_embeddings,
            cooccurrence_iterations=cooccurrence_iterations,
            cooccurrence_window_size=cooccurrence_window_size,
            number_of_epochs=number_of_epochs,
            number_of_edges_per_mini_batch=number_of_edges_per_mini_batch,
            sample_only_edges_with_heterogeneous_node_types=sample_only_edges_with_heterogeneous_node_types,
            learning_rate=learning_rate,
            first_order_decay_factor=first_order_decay_factor,
            second_order_decay_factor=second_order_decay_factor,
            avoid_false_negatives=avoid_false_negatives,
            use_scale_free_distribution=use_scale_free_distribution,
        )
        self._verbose = verbose
        self._model = models.EdgePredictionPerceptron(
            **self._model_kwargs,
            random_state=random_state
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        return dict(
            **super().parameters(),
            **self._model_kwargs
        )

    def clone(self) -> "PerceptronEdgePrediction":
        return PerceptronEdgePrediction(**self.parameters())

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(
            number_of_epochs=1,
            edge_features="Degree"
        )

    def _fit(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[List[np.ndarray]] = None,
    ):
        """Run fitting on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run predictions on.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[List[np.ndarray]] = None
            The node features to use.
        node_type_features: Optional[List[np.ndarray]] = None
            The node type features to use.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to use.
        edge_features: Optional[List[np.ndarray]] = None
            The edge features to use.
        """
        if edge_type_features is not None and len(edge_type_features) > 0:
            raise NotImplementedError(
                "Edge type features are not yet supported by "
                f"the model {self.model_name()} from {self.library_name()} library "
                f"for the task of {self.task_name()}. "
                f"Received object of type {type(edge_type_features)}."
            )
        
        if edge_features is not None and len(edge_features) > 0:
            raise NotImplementedError(
                "Edge features are not yet supported by "
                f"the model {self.model_name()} from {self.library_name()} library "
                f"for the task of {self.task_name()}. "
                f"Received object of type {type(edge_features)}."
            )

        new_node_features = []
        if node_features is not None:
            for node_feature in node_features:
                if not node_feature.data.c_contiguous:
                    node_feature = np.ascontiguousarray(node_feature)
                new_node_features.append(node_feature)

        if node_type_features is not None:
            for node_type_feature in node_type_features:
                node_transformer = NodeTransformer(
                    aligned_mapping=True
                )
                node_transformer.fit(
                    node_type_feature=node_type_feature
                )
                node_feature = node_transformer.transform(
                    node_types=graph
                )
                if not node_feature.data.c_contiguous:
                    node_feature = np.ascontiguousarray(node_feature)
                new_node_features.append(node_feature)

        self._model.fit(
            graph=graph,
            node_features=new_node_features,
            verbose=self._verbose,
            support=support
        )

    def _predict(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[List[np.ndarray]] = None,
    ) -> np.ndarray:
        """Run prediction on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run predictions on.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[List[np.ndarray]]
            The node features to use.
        node_type_features: Optional[List[np.ndarray]] = None
            The node type features to use.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to use.
        edge_features: Optional[List[np.ndarray]] = None
            The edge features to use.
        """
        return self._predict_proba(
            graph=graph,
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_features=edge_features
        ) > 0.5

    def _predict_proba(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[List[np.ndarray]] = None,
        node_type_features: Optional[List[np.ndarray]] = None,
        edge_type_features: Optional[List[np.ndarray]] = None,
        edge_features: Optional[List[np.ndarray]] = None,
    ) -> np.ndarray:
        """Run prediction on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run predictions on.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[List[np.ndarray]] = None
            The node features to use.
        node_type_features: Optional[List[np.ndarray]] = None
            The node type features to use.
        edge_type_features: Optional[List[np.ndarray]] = None
            The edge type features to use.
        edge_features: Optional[List[np.ndarray]] = None
            The edge features to use.
        """
        new_node_features = []
        for node_feature in node_features:
            if not node_feature.data.c_contiguous:
                node_feature = np.ascontiguousarray(node_feature)
            new_node_features.append(node_feature)
        
        for node_type_feature in node_type_features:
            node_transformer = NodeTransformer(
                aligned_mapping=True
            )
            node_transformer.fit(
                node_type_feature=node_type_feature
            )
            node_feature = node_transformer.transform(
                node_types=graph
            )
            if not node_feature.data.c_contiguous:
                node_feature = np.ascontiguousarray(node_feature)
            new_node_features.append(node_feature)

        return self._model.predict(
            graph=graph,
            node_features=new_node_features,
            support=support
        )

    @classmethod
    def requires_node_types(cls) -> bool:
        """Returns whether the model requires node types."""
        return False

    @classmethod
    def can_use_node_types(cls) -> bool:
        """Returns whether the model can optionally use node types."""
        return True
    
    @classmethod
    def requires_edge_types(cls) -> bool:
        """Returns whether the model requires edge types."""
        return False
    
    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return True
    
    @classmethod
    def can_use_edge_weights(cls) -> bool:
        """Returns whether the model can optionally use edge weights."""
        return False

    @classmethod
    def model_name(cls) -> str:
        return "Perceptron"

    @classmethod
    def library_name(cls) -> str:
        return "Ensmallen"

    @classmethod
    def load(cls, path: str):
        """Load a saved version of the model from the provided path.

        Parameters
        -------------------
        path: str
            Path from where to load the model.
        """
        data = compress_json.load(path)
        model = PerceptronEdgePrediction(**data["parameters"])
        model._model = models.EdgePredictionPerceptron.loads(
            json.dumps(data["inner_model"])
        )
        for key, value in data["metadata"].items():
            model.__setattr__(key, value)
        return model

    def dumps(self) -> Dict[str, Any]:
        """Dumps the current model as dictionary."""
        return dict(
            parameters=self.parameters(),
            inner_model=json.loads(self._model.dumps()),
            metadata=dict(
                _fitting_was_executed=self._fitting_was_executed
            )
        )

    def dump(self, path: str):
        """Dump the current model at the provided path.

        Parameters
        -------------------
        path: str
            Path from where to dump the model.
        """
        if ".json" in path:
            compress_json.dump(
                self.dumps(),
                path
            )
        if ".pkl" in path:
            compress_pickle.dump(
                self.dumps(),
                path
            )

        raise NotImplementedError(
            "The provided path does not have a valid extension. "
            "Please use either .json or .pkl, followed by the eventual "
            "desired compression algorithm extension."
        )

    @classmethod
    def can_use_edge_type_features(cls) -> bool:
        return False
    
    @classmethod
    def can_use_edge_features(cls) -> bool:
        return False
    
    


"""Module providing abstract edge prediction model."""
import warnings
from typing import Optional, Union, List, Dict, Any, Tuple, Iterator, Type
import pandas as pd
import numpy as np
import math
from ensmallen import Graph
from tqdm.auto import tqdm
from embiggen.utils import AbstractEdgeFeature
from embiggen.utils.abstract_models import (
    AbstractClassifierModel,
    abstract_class,
    format_list,
)


@abstract_class
class AbstractEdgePredictionModel(AbstractClassifierModel):
    """Class defining an abstract edge prediction model."""

    @classmethod
    def task_name(cls) -> str:
        """Returns name of the task this model is used for."""
        return "Edge Prediction"

    def is_binary_prediction_task(self) -> bool:
        """Returns whether the model was fit on a binary prediction task."""
        # Edge prediction is always a binary prediction task.
        return True

    @classmethod
    def is_topological(cls) -> bool:
        return True

    @classmethod
    def get_available_evaluation_schemas(cls) -> List[str]:
        """Returns available evaluation schemas for this task."""
        return ["Connected Monte Carlo", "Monte Carlo", "Kfold"]

    @classmethod
    def edge_features_check(
        cls,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ],
    ) -> Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]]:
        """Check the provided edge features."""
        if edge_features is None:
            edge_features = []
        if not isinstance(edge_features, list):
            edge_features = [edge_features]
        for edge_feature in edge_features:
            if not issubclass(type(edge_feature), AbstractEdgeFeature):
                raise NotImplementedError(
                    "Currently, we solely support edge features that are subclasses of AbstractEdgeFeature. "
                    "This is because most commonly, it is not possible to precompute edge features for all "
                    "possible edges of a complete graph and thus, we need to compute them on the fly. "
                    "To do so, we need a common interface that allows us to query the edge features on "
                    "demand, lazily, hence avoiding unsustainable memory peaks."
                    f"You have provided an egde feature of type {type(edge_feature)}, which is not a subclass of AbstractEdgeFeature."
                )
        return edge_features

    @classmethod
    def split_graph_following_evaluation_schema(
        cls,
        graph: Graph,
        evaluation_schema: str,
        random_state: int,
        holdout_number: int,
        number_of_holdouts: int,
        **holdouts_kwargs: Dict[str, Any],
    ) -> Tuple[Graph]:
        """Return train and test graphs tuple following the provided evaluation schema.

        Parameters
        ----------------------
        graph: Graph
            The graph to split.
        evaluation_schema: str
            The evaluation schema to follow.
        random_state: int
            The random state for the evaluation
        holdout_number: int
            The current holdout number.
        number_of_holdouts: int
            The total number of holdouts.
        holdouts_kwargs: Dict[str, Any]
            The kwargs to be forwarded to the holdout method.
        """
        if evaluation_schema == "Connected Monte Carlo":
            return graph.connected_holdout(
                **holdouts_kwargs,
                random_state=random_state + holdout_number,
                verbose=False,
            )
        if evaluation_schema == "Monte Carlo":
            return graph.random_holdout(
                **holdouts_kwargs,
                random_state=random_state + holdout_number,
                verbose=False,
            )
        if evaluation_schema == "Kfold":
            return graph.get_edge_prediction_kfold(
                k=number_of_holdouts,
                k_index=holdout_number,
                random_state=random_state,
                verbose=False,
            )
        super().split_graph_following_evaluation_schema(
            graph=graph,
            evaluation_schema=evaluation_schema,
            random_state=random_state,
            holdout_number=holdout_number,
            number_of_holdouts=number_of_holdouts,
            **holdouts_kwargs,
        )

    @staticmethod
    def __iterate_negative_graphs(
        graph: Graph,
        train: Graph,
        test: Graph,
        support: Graph,
        subgraph_of_interest: Optional[Graph],
        random_state: int,
        verbose: bool,
        validation_sample_only_edges_with_heterogeneous_node_types: bool,
        source_node_types_names: Optional[List[str]],
        destination_node_types_names: Optional[List[str]],
        source_edge_types_names: Optional[List[str]],
        destination_edge_types_names: Optional[List[str]],
        source_nodes_prefixes: Optional[List[str]],
        destination_nodes_prefixes: Optional[List[str]],
        validation_unbalance_rates: Tuple[float],
        use_scale_free_distribution: bool,
    ) -> Iterator[Tuple[Graph]]:
        """Return iterator over the negative graphs for evaluation."""
        if subgraph_of_interest is None:
            sampler_graph = graph
        else:
            sampler_graph = subgraph_of_interest

        if not use_scale_free_distribution:
            warnings.warn(
                "Please do be advised that you have DISABLED the use of scale free sampling "
                "for the negative edges for the EVALUATION (not the training) "
                "of a model. This is a POOR CHOICE as it will introduce a positive bias "
                "as edges sampled uniformely have a significantly different node degree "
                "distribution than the positive edges in the graph, and are therefore much easier "
                "to predict. The only case where it makes sense to use this parameter is when "
                "evaluating how strongly this bias would have affected your task. "
                "DO NOT USE THIS CONFIGURATION IN ANY OTHER USE CASE."
            )

        train_size = train.get_number_of_edges() / (
            train.get_number_of_edges() + test.get_number_of_edges()
        )

        return (
            sampler_graph.sample_negative_graph(
                number_of_negative_samples=int(
                    math.ceil(sampler_graph.get_number_of_edges() * unbalance_rate)
                ),
                random_state=random_state * (i + 1),
                sample_only_edges_with_heterogeneous_node_types=validation_sample_only_edges_with_heterogeneous_node_types,
                use_scale_free_distribution=use_scale_free_distribution,
                source_node_types_names=source_node_types_names,
                destination_node_types_names=destination_node_types_names,
                source_edge_types_names=source_edge_types_names,
                destination_edge_types_names=destination_edge_types_names,
                source_nodes_prefixes=source_nodes_prefixes,
                destination_nodes_prefixes=destination_nodes_prefixes,
                support=support,
                graph_to_avoid=graph,
            ).random_holdout(
                train_size=train_size,
                random_state=random_state,
                verbose=False,
            )
            for i, unbalance_rate in tqdm(
                enumerate(validation_unbalance_rates),
                disable=not verbose or len(validation_unbalance_rates) == 1,
                total=len(validation_unbalance_rates),
                leave=False,
                dynamic_ncols=True,
                desc="Building negative graphs for evaluation",
            )
        )

    @classmethod
    def _prepare_evaluation(
        cls,
        graph: Graph,
        train: Graph,
        test: Graph,
        support: Optional[Graph] = None,
        subgraph_of_interest: Optional[Graph] = None,
        random_state: int = 42,
        verbose: bool = True,
        validation_sample_only_edges_with_heterogeneous_node_types: bool = False,
        source_node_types_names: Optional[List[str]] = None,
        destination_node_types_names: Optional[List[str]] = None,
        source_edge_types_names: Optional[List[str]] = None,
        destination_edge_types_names: Optional[List[str]] = None,
        source_nodes_prefixes: Optional[List[str]] = None,
        destination_nodes_prefixes: Optional[List[str]] = None,
        validation_unbalance_rates: Tuple[float] = (1.0,),
        use_scale_free_distribution: bool = True,
    ) -> Dict[str, Any]:
        """Return additional custom parameters for the current holdout."""
        return dict(
            negative_graphs=list(
                cls.__iterate_negative_graphs(
                    graph=graph,
                    train=train,
                    test=test,
                    support=support,
                    subgraph_of_interest=subgraph_of_interest,
                    random_state=random_state,
                    verbose=verbose,
                    validation_sample_only_edges_with_heterogeneous_node_types=validation_sample_only_edges_with_heterogeneous_node_types,
                    source_node_types_names=source_node_types_names,
                    destination_node_types_names=destination_node_types_names,
                    source_edge_types_names=source_edge_types_names,
                    destination_edge_types_names=destination_edge_types_names,
                    source_nodes_prefixes=source_nodes_prefixes,
                    destination_nodes_prefixes=destination_nodes_prefixes,
                    validation_unbalance_rates=validation_unbalance_rates,
                    use_scale_free_distribution=use_scale_free_distribution,
                )
            )
        )

    def _evaluate(
        self,
        graph: Graph,
        train: Graph,
        test: Graph,
        support: Graph,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[str, pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[str, pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[str, pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[str, pd.DataFrame, np.ndarray]]]
        ] = None,
        subgraph_of_interest: Optional[Graph] = None,
        random_state: int = 42,
        verbose: bool = True,
        negative_graphs: Optional[List[Tuple[Graph]]] = None,
        validation_sample_only_edges_with_heterogeneous_node_types: bool = False,
        source_node_types_names: Optional[List[str]] = None,
        destination_node_types_names: Optional[List[str]] = None,
        source_edge_types_names: Optional[List[str]] = None,
        destination_edge_types_names: Optional[List[str]] = None,
        source_nodes_prefixes: Optional[List[str]] = None,
        destination_nodes_prefixes: Optional[List[str]] = None,
        validation_unbalance_rates: Tuple[float] = (1.0,),
        use_scale_free_distribution: bool = True,
    ) -> List[Dict[str, Any]]:
        """Return model evaluation on the provided graphs."""
        performance = []

        train_size = (
            train.get_number_of_directed_edges() / graph.get_number_of_directed_edges()
        )

        train_predict_proba = self.predict_proba(
            train,
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
        )

        test_predict_proba = self.predict_proba(
            test,
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
        )

        negative_graph_iterator = (
            self.__iterate_negative_graphs(
                graph=graph,
                train=train,
                test=test,
                support=support,
                subgraph_of_interest=subgraph_of_interest,
                random_state=random_state,
                verbose=verbose,
                validation_sample_only_edges_with_heterogeneous_node_types=validation_sample_only_edges_with_heterogeneous_node_types,
                source_node_types_names=source_node_types_names,
                destination_node_types_names=destination_node_types_names,
                source_edge_types_names=source_edge_types_names,
                destination_edge_types_names=destination_edge_types_names,
                source_nodes_prefixes=source_nodes_prefixes,
                destination_nodes_prefixes=destination_nodes_prefixes,
                validation_unbalance_rates=validation_unbalance_rates,
                use_scale_free_distribution=use_scale_free_distribution,
            )
            if negative_graphs is None
            else negative_graphs
        )

        for unbalance_rate, (negative_train, negative_test) in tqdm(
            zip(validation_unbalance_rates, negative_graph_iterator),
            disable=not verbose or len(validation_unbalance_rates) == 1,
            total=len(validation_unbalance_rates),
            leave=False,
            dynamic_ncols=True,
            desc=f"Evaluating on unbalances",
        ):
            for evaluation_mode, (existent_predict_proba, non_existent_graph) in (
                ("train", (train_predict_proba, negative_train)),
                ("test", (test_predict_proba, negative_test)),
            ):
                non_existent_predict_proba = self.predict_proba(
                    non_existent_graph,
                    support=support,
                    node_features=node_features,
                    node_type_features=node_type_features,
                    edge_type_features=edge_type_features,
                    edge_features=edge_features,
                )

                if (
                    len(non_existent_predict_proba.shape) > 1
                    and non_existent_predict_proba.shape[1] > 1
                ):
                    non_existent_predict_proba = non_existent_predict_proba[:, 1]

                predict_proba = np.concatenate(
                    (existent_predict_proba, non_existent_predict_proba)
                )

                labels = np.concatenate(
                    (
                        np.ones_like(existent_predict_proba, dtype=bool),
                        np.zeros_like(non_existent_predict_proba, dtype=bool),
                    )
                )

                performance.append(
                    {
                        "evaluation_mode": evaluation_mode,
                        "train_size": train_size,
                        "validation_unbalance_rate": unbalance_rate,
                        "use_scale_free_distribution": use_scale_free_distribution,
                        "validation_sample_only_edges_with_heterogeneous_node_types": validation_sample_only_edges_with_heterogeneous_node_types,
                        **self.evaluate_predictions(
                            labels,
                            predict_proba,
                        ),
                        **self.evaluate_prediction_probabilities(
                            labels,
                            predict_proba,
                        ),
                    }
                )

        return performance

    def predict(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run predictions on.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        if graph.has_edges():
            predictions = (
                super()
                .predict(
                    graph,
                    support=support,
                    node_features=node_features,
                    node_type_features=node_type_features,
                    edge_type_features=edge_type_features,
                    edge_features=self.edge_features_check(edge_features),
                )
                .flatten()
            )
        else:
            predictions = np.array([])

        if return_predictions_dataframe:
            predictions = pd.DataFrame(
                {
                    "predictions": predictions,
                    "sources": graph.get_source_names(directed=True)
                    if return_node_names
                    else graph.get_directed_source_node_ids(),
                    "destinations": graph.get_destination_names(directed=True)
                    if return_node_names
                    else graph.get_directed_destination_node_ids(),
                    **(
                        {
                            "edge_types": graph.get_directed_edge_type_names()
                            if return_edge_type_names
                            else graph.get_directed_edge_type_ids()
                        }
                        if graph.has_edge_types()
                        else {}
                    ),
                },
            )

        return predictions

    def predict_bipartite_graph_from_edge_node_ids(
        self,
        graph: Graph,
        source_node_ids: List[int],
        destination_node_ids: List[int],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_ids: List[int]
            The source nodes of the bipartite graph.
        destination_node_ids: List[int]
            The destination nodes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict(
            graph.build_bipartite_graph_from_edge_node_ids(
                source_node_ids=source_node_ids,
                destination_node_ids=destination_node_ids,
                directed=True,
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_bipartite_graph_from_edge_node_names(
        self,
        graph: Graph,
        source_node_names: List[str],
        destination_node_names: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_names: List[str]
            The source nodes of the bipartite graph.
        destination_node_names: List[str]
            The destination nodes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict(
            graph.build_bipartite_graph_from_edge_node_names(
                source_node_names=source_node_names,
                destination_node_names=destination_node_names,
                directed=True,
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_bipartite_graph_from_edge_node_prefixes(
        self,
        graph: Graph,
        source_node_prefixes: List[str],
        destination_node_prefixes: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_prefixes: List[str]
            The source node prefixes of the bipartite graph.
        destination_node_prefixes: List[str]
            The destination node prefixes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict(
            graph.build_bipartite_graph_from_edge_node_prefixes(
                source_node_prefixes=source_node_prefixes,
                destination_node_prefixes=destination_node_prefixes,
                directed=True,
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_bipartite_graph_from_edge_node_types(
        self,
        graph: Graph,
        source_node_types: List[str],
        destination_node_types: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_types: List[str]
            The source node prefixes of the bipartite graph.
        destination_node_types: List[str]
            The destination node prefixes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict(
            graph.build_bipartite_graph_from_edge_node_types(
                source_node_types=source_node_types,
                destination_node_types=destination_node_types,
                directed=True,
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_clique_graph_from_node_ids(
        self,
        graph: Graph,
        node_ids: List[int],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_ids: List[int]
            The nodes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict(
            graph.build_clique_graph_from_node_ids(node_ids=node_ids, directed=True),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_clique_graph_from_node_names(
        self,
        graph: Graph,
        node_names: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_names: List[str]
            The nodes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict(
            graph.build_clique_graph_from_node_names(
                node_names=node_names, directed=True
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_clique_graph_from_node_prefixes(
        self,
        graph: Graph,
        node_prefixes: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_prefixes: List[str]
            The node prefixes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict(
            graph.build_clique_graph_from_node_prefixes(
                node_prefixes=node_prefixes, directed=True
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_clique_graph_from_node_type_names(
        self,
        graph: Graph,
        node_type_names: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_type_names: List[str]
            The node prefixes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict(
            graph.build_clique_graph_from_node_type_names(
                node_type_names=node_type_names, directed=True
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_proba(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run predictions on.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        predictions = (
            super()
            .predict_proba(
                graph,
                support=support,
                node_features=node_features,
                node_type_features=node_type_features,
                edge_type_features=edge_type_features,
                edge_features=self.edge_features_check(edge_features),
            )
            .flatten()
        )

        if np.isnan(predictions).any():
            raise ValueError("There are NaN values in the predicted probabilities!")

        if return_predictions_dataframe:
            predictions = pd.DataFrame(
                {
                    "predictions": predictions,
                    "sources": graph.get_source_names(directed=True)
                    if return_node_names
                    else graph.get_directed_source_node_ids(),
                    "destinations": graph.get_destination_names(directed=True)
                    if return_node_names
                    else graph.get_directed_destination_node_ids(),
                    **(
                        {
                            "edge_types": graph.get_directed_edge_type_names()
                            if return_edge_type_names
                            else graph.get_directed_edge_type_ids()
                        }
                        if graph.has_edge_types()
                        else {}
                    ),
                },
            )

        return predictions

    def predict_proba_bipartite_graph_from_edge_node_ids(
        self,
        graph: Graph,
        source_node_ids: List[int],
        destination_node_ids: List[int],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_ids: List[int]
            The source nodes of the bipartite graph.
        destination_node_ids: List[int]
            The destination nodes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict_proba(
            graph.build_bipartite_graph_from_edge_node_ids(
                source_node_ids=source_node_ids,
                destination_node_ids=destination_node_ids,
                directed=True,
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_proba_bipartite_graph_from_edge_node_names(
        self,
        graph: Graph,
        source_node_names: List[str],
        destination_node_names: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_names: List[str]
            The source nodes of the bipartite graph.
        destination_node_names: List[str]
            The destination nodes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict_proba(
            graph.build_bipartite_graph_from_edge_node_names(
                source_node_names=source_node_names,
                destination_node_names=destination_node_names,
                directed=True,
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_proba_bipartite_graph_from_edge_node_prefixes(
        self,
        graph: Graph,
        source_node_prefixes: List[str],
        destination_node_prefixes: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_prefixes: List[str]
            The source node prefixes of the bipartite graph.
        destination_node_prefixes: List[str]
            The destination node prefixes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict_proba(
            graph.build_bipartite_graph_from_edge_node_prefixes(
                source_node_prefixes=source_node_prefixes,
                destination_node_prefixes=destination_node_prefixes,
                directed=True,
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_proba_bipartite_graph_from_edge_node_types(
        self,
        graph: Graph,
        source_node_types: List[str],
        destination_node_types: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        source_node_types: List[str]
            The source node prefixes of the bipartite graph.
        destination_node_types: List[str]
            The destination node prefixes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict_proba(
            graph.build_bipartite_graph_from_edge_node_types(
                source_node_types=source_node_types,
                destination_node_types=destination_node_types,
                directed=True,
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_proba_clique_graph_from_node_ids(
        self,
        graph: Graph,
        node_ids: List[int],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_ids: List[int]
            The nodes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict_proba(
            graph.build_clique_graph_from_node_ids(node_ids=node_ids, directed=True),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_proba_clique_graph_from_node_names(
        self,
        graph: Graph,
        node_names: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_names: List[str]
            The nodes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict_proba(
            graph.build_clique_graph_from_node_names(
                node_names=node_names, directed=True
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_proba_clique_graph_from_node_prefixes(
        self,
        graph: Graph,
        node_prefixes: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_prefixes: List[str]
            The node prefixes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict_proba(
            graph.build_clique_graph_from_node_prefixes(
                node_prefixes=node_prefixes, directed=True
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def predict_proba_clique_graph_from_node_type_names(
        self,
        graph: Graph,
        node_type_names: List[str],
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
        return_predictions_dataframe: bool = False,
        return_edge_type_names: bool = True,
        return_node_names: bool = True,
    ) -> Union[np.ndarray, pd.DataFrame]:
        """Execute predictions probabilities on the provided graph bipartite portion.

        Parameters
        --------------------
        graph: Graph
            The graph from which to extract the edges.
        node_type_names: List[str]
            The node prefixes of the bipartite graph.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        return_predictions_dataframe: bool = False
            Whether to return a pandas DataFrame, which as indices has the node IDs.
            By default, a numpy array with the predictions is returned as it weights much less.
        return_edge_type_names: bool = True
            Whether to return edge type names when returning the prediction DataFrame.
            This value is ignored when either the values to be returned to the user
            are not requested to be a DataFrame or the graph does not have edge types.
        return_node_names: bool = True
            Whether to return node names when returning the prediction DataFrame.
            This value is ignored when the values to be returned the user has not
            requested for a prediction dataframe to be returned.
        """
        return self.predict_proba(
            graph.build_clique_graph_from_node_type_names(
                node_type_names=node_type_names, directed=True
            ),
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=edge_features,
            return_predictions_dataframe=return_predictions_dataframe,
            return_edge_type_names=return_edge_type_names,
            return_node_names=return_node_names,
        )

    def fit(
        self,
        graph: Graph,
        support: Optional[Graph] = None,
        node_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        node_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_features: Optional[
            Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]
        ] = None,
    ):
        """Execute fitting on the provided graph.

        Parameters
        --------------------
        graph: Graph
            The graph to run predictions on.
        support: Optional[Graph] = None
            The graph describiding the topological structure that
            includes also the above graph. This parameter
            is mostly useful for topological classifiers
            such as Graph Convolutional Networks.
        node_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node features to use.
        node_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            The node type features to use.
        edge_features: Optional[Union[Type[AbstractEdgeFeature], List[Type[AbstractEdgeFeature]]]] = None
            The edge features to use.
        """
        super().fit(
            graph=graph,
            support=support,
            node_features=node_features,
            node_type_features=node_type_features,
            edge_type_features=edge_type_features,
            edge_features=self.edge_features_check(edge_features),
        )

    def is_using_node_types(self) -> bool:
        """Whether the current model is using node types."""
        return self._is_using_node_type_features or self.requires_node_types()

    def is_using_edge_types(self) -> bool:
        """Whether the current model is using edge types."""
        return self._is_using_edge_type_features or self.requires_edge_types()

    @classmethod
    def task_involves_edge_weights(cls) -> bool:
        """Returns whether the model task involves edge weights."""
        return False

    @classmethod
    def task_involves_edge_types(cls) -> bool:
        """Returns whether the model task involves edge types."""
        return False

    @classmethod
    def task_involves_node_types(cls) -> bool:
        """Returns whether the model task involves node types."""
        return False

    @classmethod
    def task_involves_topology(cls) -> bool:
        """Returns whether the model task involves topology."""
        return True

    @classmethod
    def supports_multilabel_prediction(cls) -> bool:
        """Returns whether the model supports multilabel prediction.

        Implementation details
        ----------------------
        The edge prediction task is, by definition, a binary prediction
        task. Therefore, this method always returns False.
        """
        return False

    def is_multilabel_prediction_task(self) -> bool:
        """Returns whether the model is a multilabel prediction task.

        Implementation details
        ----------------------
        The edge prediction task is, by definition, a binary prediction
        task. Therefore, this method always returns False.
        """
        return False

    @classmethod
    def can_use_node_type_features(cls) -> bool:
        """Returns whether the model can use node type features."""
        return True


import pytest
import inspect
import embiggen
import warnings

def check_alls(module):
    error_msg = ""

    # Check for module documentation of __init__ files
    if inspect.getfile(module).endswith("__init__.py"):
        doc = getattr(module, "__doc__", None)
        if doc is None or module.__doc__.strip() == "":
            warnings.warn("The file '{}' doesn't have documentation!".format(inspect.getfile(module)))

    alls = getattr(module, "__all__", None)
    # warn about the missing __all__ and return
    if alls is None:
        if inspect.getfile(module).endswith("__init__.py"):
            warnings.warn("The module '{}' doesn't have an __all__".format(module.__name__))
    else:
        # Check that all the values in all resolves
        missings = set()
        for sub in alls:
            if sub not in dir(module):
                missings.add(sub)

        if len(missings) > 0:
            error_msg += "in the __all__ of module '{}' the following values '{}' cannot be imported\n".format(
                module.__name__, missings,
            )

        # Check for duplicates
        if len(alls) != len(set(alls)):
            duplicates = list(sorted({
                x
                for x in alls
                if alls.count(x) > 1
            }))
            error_msg += "is in the __all__ of module '{}' the following keys are duplicated '{}'\n".format(
                module.__name__, duplicates,
            )

        if len(error_msg) > 0:
            raise AssertionError(error_msg)

    # Recursive step for submodules (even if not exported in __all__)
    for value in vars(module).values():
        # only check modules inside embiggen
        mod = inspect.getmodule(value)
        if mod is None or "embiggen" not in mod.__name__:
            continue

        if inspect.ismodule(value):
            check_alls(value)

def test_alls():
    check_alls(embiggen)



"""GraphTransformer class to convert graphs to edge embeddings."""
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from ensmallen import Graph  # pylint: disable=no-name-in-module

from embiggen.embedding_transformers.edge_transformer import EdgeTransformer


class GraphTransformer:
    """GraphTransformer class to convert graphs to edge embeddings."""

    def __init__(
        self,
        methods: Union[List[str], str] = "Hadamard",
        aligned_mapping: bool = False,
        include_both_undirected_edges: bool = True,
    ):
        """Create new GraphTransformer object.

        Parameters
        ------------------------
        methods: Union[List[str], str] = "Hadamard"
            Method to use for the edge embedding.
            If multiple edge embedding are provided, they
            will be Concatenated and fed to the model.
            The supported edge embedding methods are:
             * Hadamard: element-wise product
             * Sum: element-wise sum
             * Average: element-wise mean
             * L1: element-wise subtraction
             * AbsoluteL1: element-wise subtraction in absolute value
             * SquaredL2: element-wise subtraction in squared value
             * L2: element-wise squared root of squared subtraction
             * Concatenate: Concatenate of source and destination node features
             * Min: element-wise minimum
             * Max: element-wise maximum
             * L2Distance: vector-wise L2 distance - this yields a scalar
             * CosineSimilarity: vector-wise cosine similarity - this yields a scalar
        aligned_mapping: bool = False
            This parameter specifies whether the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        include_both_undirected_edges: bool = True
            Whether to include both undirected edges when parsing an undirected
            graph, that is both the edge from source to destination and the edge
            from destination to source. While both edges should be included when
            training a model, as the model should learn about these simmetries
            in the graph, these edges are not necessary in the context of visualizations
            where they create redoundancy.
        """
        self._transformer = EdgeTransformer(
            methods=methods,
            aligned_mapping=aligned_mapping,
        )
        self._include_both_undirected_edges = include_both_undirected_edges
        self._aligned_mapping = aligned_mapping

    def fit(
        self,
        node_feature: Union[
            pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]
        ],
        node_type_feature: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
        edge_type_features: Optional[
            Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
        ] = None,
    ):
        """Fit the model.

        Parameters
        -------------------------
        node_feature: Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]
            Node feature to use to fit the transformer.
        node_type_feature: Optional[Union[pd.DataFrame, np.ndarray, List[Union[pd.DataFrame, np.ndarray]]]] = None
            Node type feature to use to fit the transformer.
        edge_type_features: Optional[Union[pd.DataFrame, np.ndarray,
                                             List[Union[pd.DataFrame, np.ndarray]]]] = None
            Edge type feature to use to fit the transformer.

        Raises
        -------------------------
        ValueError
            If the given method is None there is no need to call the fit method.
        """
        self._transformer.fit(
            node_feature=node_feature,
            node_type_feature=node_type_feature,
            edge_type_features=edge_type_features,
        )

    def has_node_type_features(self) -> bool:
        """Return whether the transformer has a node type feature."""
        return self._transformer.has_node_type_features()

    def has_edge_type_features(self) -> bool:
        """Return whether the transformer has a edge type feature."""
        return self._transformer.has_edge_type_features()

    def is_aligned_mapping(self) -> bool:
        """Return whether the transformer has a aligned mapping."""
        return self._transformer.is_aligned_mapping()

    def transform(
        self,
        graph: Union[Graph, np.ndarray, List[List[str]], List[List[int]]],
        node_types: Optional[
            Union[Graph, List[Optional[List[str]]], List[Optional[List[int]]]]
        ] = None,
        edge_types: Optional[Union[Graph, List[str], List[int], np.ndarray]] = None,
        edge_features: Optional[Union[np.ndarray, List[np.ndarray]]] = None,
    ) -> np.ndarray:
        """Return edge embedding for given graph using provided method.

        Parameters
        --------------------------
        graph: Union[Graph, np.ndarray, List[List[str]], List[List[int]]],
            The graph whose edges are to embed.
            It can either be an Graph or a list of lists of edges.
        node_types: Optional[Union[Graph, List[Optional[List[str]]], List[Optional[List[int]]]]] = None,
            List of node types whose embedding is to be returned.
            This can be either a list of strings, or a graph, or if the
            aligned_mapping is setted, then this methods also accepts
            a list of ints.
        edge_features: Optional[Union[np.ndarray, List[np.ndarray]]] = None
            Optional edge features to be used as input Concatenated
            to the obtained edge embedding. The shape must be equal
            to the number of directed edges in the provided graph.

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Numpy array of embeddings.
        """
        if isinstance(graph, Graph):
            if self._aligned_mapping:
                if graph.is_directed() or self._include_both_undirected_edges:
                    edge_node_ids = (
                        graph.get_directed_source_node_ids(),
                        graph.get_directed_destination_node_ids(),
                    )
                else:
                    edge_node_ids = (
                        graph.get_source_node_ids(directed=False),
                        graph.get_destination_node_ids(directed=False),
                    )
            else:
                edge_node_ids = graph.get_directed_edge_node_names()
        else:
            edge_node_ids = graph

        if isinstance(edge_node_ids, List):
            edge_node_ids = np.array(edge_node_ids)
        if (
            isinstance(edge_node_ids, tuple)
            and len(edge_node_ids) == 2
            and all(isinstance(e, np.ndarray) for e in edge_node_ids)
        ):
            if (
                len(edge_node_ids[0].shape) != 1
                or len(edge_node_ids[1].shape) != 1
                or edge_node_ids[0].shape[0] == 0
                or edge_node_ids[1].shape[0] == 0
                or edge_node_ids[0].shape[0] != edge_node_ids[1].shape[0]
            ):
                raise ValueError(
                    "When providing a tuple of numpy arrays containing the source and destination "
                    "node IDs, we expect to receive two arrays both with shape "
                    "with shape (number of edges,). "
                    f"The ones you have provided have shapes {edge_node_ids[0].shape} "
                    f"and {edge_node_ids[1].shape}."
                )
            sources = edge_node_ids[0]
            destinations = edge_node_ids[1]
        elif isinstance(edge_node_ids, np.ndarray):
            if (
                len(edge_node_ids.shape) != 2
                or edge_node_ids.shape[1] != 2
                or edge_node_ids.shape[0] == 0
            ):
                raise ValueError(
                    "When providing a numpy array containing the source and destination "
                    "node IDs representing the graph edges, we expect to receive an array "
                    f"with shape (number of edges, 2). The one you have provided has shape {edge_node_ids.shape}."
                )
            sources = edge_node_ids[:, 0]
            destinations = edge_node_ids[:, 1]

        if node_types is not None and self.has_node_type_features():
            if isinstance(node_types, Graph):
                if self._aligned_mapping:
                    source_node_types = [
                        node_types.get_node_type_ids_from_node_id(src)
                        for src in sources
                    ]
                    destination_node_types = [
                        node_types.get_node_type_ids_from_node_id(dst)
                        for dst in destinations
                    ]
                else:
                    source_node_types = [
                        node_types.get_node_type_names_from_node_name(src)
                        for src in sources
                    ]
                    destination_node_types = [
                        node_types.get_node_type_names_from_node_name(dst)
                        for dst in destinations
                    ]
            else:
                source_node_types, destination_node_types = node_types
        else:
            source_node_types = None
            destination_node_types = None

        assert (source_node_types is not None) == self.has_node_type_features()
        assert (destination_node_types is not None) == self.has_node_type_features()

        if isinstance(edge_types, Graph):
            edge_types.must_not_contain_unknown_edge_types()
            edge_types.must_not_be_multigraph()
            if not self.has_edge_type_features():
                raise ValueError(
                    "While the provided graph has edge types, "
                    "no edge features were provided to the graph transformer"
                )
            if self.is_aligned_mapping():
                if edge_types.is_directed() or self._include_both_undirected_edges:
                    edge_types = edge_types.get_imputed_directed_edge_type_ids(
                        imputation_edge_type_id=0
                    )
                else:
                    edge_types = edge_types.get_imputed_upper_triangular_edge_type_ids(
                        imputation_edge_type_id=0
                    )
            else:
                if edge_types.is_directed() or self._include_both_undirected_edges:
                    edge_types = edge_types.get_directed_edge_type_names()
                else:
                    edge_types = edge_types.get_upper_triangular_edge_type_names()

        assert (edge_types is not None) == self.has_edge_type_features()

        return self._transformer.transform(
            sources,
            destinations,
            source_node_types=source_node_types,
            destination_node_types=destination_node_types,
            edge_types=edge_types,
            edge_features=edge_features,
        )



"""Submodule with utils for interface with Sklearn models."""
from sklearn.base import ClassifierMixin
from sklearn.linear_model._base import LinearClassifierMixin


def is_sklearn_classifier_model(candidate_model) -> bool:
    """Returns whether a given object is a Sklearn classifier model."""
    return issubclass(
        type(candidate_model), (ClassifierMixin, LinearClassifierMixin)
    ) and all(
        hasattr(candidate_model, method_name)
        for method_name in (
            "fit",
        )
    ) and any(
        hasattr(candidate_model, method_name)
        for method_name in (
            "predict_proba",
            "predict",
        )
    )


def must_be_an_sklearn_classifier_model(candidate_model):
    """Raises an exception if the provided object is not an sklearn classifier model."""
    if not is_sklearn_classifier_model(candidate_model):
        raise ValueError(
            f"The provided object of type {type(candidate_model)} is not a valid sklearn classifier model "
            "that can be adapted for this class as it is not a subclass of `ClassifierMixin` "
            "nor `LinearClassifierMixin`."
        )



"""Submodule providing auto model stub for non-mandatory modules."""
from typing import Type, List, Union, Dict, Any
from embiggen.utils.abstract_models.abstract_model import AbstractModel
from embiggen.utils.abstract_models.list_formatting import format_list


def get_model_or_stub(
    frame,
    module_library_names: Union[str, List[str]],
    formatted_library_name: str,
    submodule_name: str,
    model_class_name: str,
    formatted_model_name: str,
    task_name: str,
    parent_class: Type[AbstractModel]
):
    """Returns either the class or a stub with helpful error messages.

    Parameters
    -------------------
    frame
        Stack frame of the main context.
    module_library_names: Union[str, List[str]]
        Name of the library dependencies to be check for availability.
    formatted_library_name: str
        The formatted name of the library for visualization pourposes.
    submodule_name: str
        Name of the submodule to load.
    model_class_name: str
        Name of the model class to load or stub.
    formatted_model_name: str
        Formatted model name to load.
    task_name: str
        Name of the task of the model.
    parent_class: Type[AbstractModel]
        Expected parent class of the model.
    """
    if not isinstance(module_library_names, list):
        module_library_names = [module_library_names]
    
    # We check that all names are actually string.
    for variable_name, variable_value in (
        ("formatted_library_name", formatted_library_name),
        ("submodule_name", submodule_name),
        ("model_class_name", model_class_name),
        *[
            ("module_library_name", module_library_name)
            for module_library_name in module_library_names
        ]
    ):
        if not isinstance(variable_value, str):
            raise ValueError(
                f"The provided parent class {variable_name} is not string. "
                "This is likely an implementation error, and should be "
                "reported to the Embiggen repository as an issue."
            )
        if variable_value == "":
            raise ValueError(
                f"The provided parent class {variable_name} is an empty string. "
                "This is likely an implementation error, and should be "
                "reported to the Embiggen repository as an issue."
            )
    # We check that the provided parent class is actually a subclass
    # of the AbstractModel interface.
    if not issubclass(parent_class, AbstractModel):
        raise ValueError(
            f"The provided parent class {parent_class} is not a child "
            "of AbstractModel. This is "
            "likely an implementation error, and should be "
            "reported to the Embiggen repository as an issue."
        )
    # We identify what module or submodule is it calling from.
    current_module_name = frame.f_back.f_locals["__name__"]
    # We try to import the required class.
    try:
        # We try to retrieve the requested model class.
        model_class = getattr(
            __import__(
                f"{current_module_name}.{submodule_name}",
                fromlist=(model_class_name,)
            ),
            model_class_name
        )
        # We check that the loaded class is effectively an
        # implementation of the expected parent class.
        if not issubclass(model_class, parent_class):
            raise ValueError(
                f"The provided model class {model_class_name} is not "
                f"an implementation of {parent_class}. This is "
                "likely an implementation error, and should be "
                "reported to the Embiggen repository as an issue."
            )
    except (ModuleNotFoundError, OSError) as exception:
        # If effectively the error is that we cannot load the desired
        # library name, we catch this and re-raise it.
        if any(
            str(exception).startswith(f"No module named '{module_library_name}")
            or module_library_name in str(exception)
            for module_library_name in module_library_names
        ):
            class StubClass(parent_class):

                def __init__(self, *args, **kwargs):
                    """Raises a useful error message about this class."""
                    self.__class__.__name__ = model_class_name
                    other_candidates = self.find_available_models(
                        self.model_name(),
                        self.task_name()
                    )
                    if other_candidates:
                        other_libraries = [
                            model_name.library_name()
                            for model_name in other_candidates
                        ]
                        other_candidates_message = (
                            "Do be advised that, while this model is not "
                            "currently available on your system "
                            "in this specific library implementation, "
                            f"the same model is implemented in {format_list(other_libraries)}. "
                            "Do be aware that different implementations may have "
                            "very different parametrizations and performance."
                        )
                    else:
                        other_candidates_message = (
                            "At this time, there is no other implementation "
                            f"of the {self.model_name()} model available for "
                            "your system."
                        )
                    raise ModuleNotFoundError(
                        (
                            f"The module {module_library_names} is not available "
                            "on your system "
                            "and therefore we cannot make available the requested "
                            f"model {self.model_name()}, as it is based on the "
                            f"{self.library_name()} library. "
                            "Please do refer to the requested library documentation "
                            f"to proceed with the installation. {other_candidates_message}"
                        )
                    )

                @staticmethod
                def library_name() -> str:
                    """Returns library name of the model."""
                    return formatted_library_name

                @classmethod
                def model_name(cls) -> str:
                    """Returns name of the model."""
                    return formatted_model_name

                @classmethod
                def requires_node_types(cls) -> str:
                    """Returns whether the model requires node types."""
                    return None

                @classmethod
                def can_use_node_types(cls) -> bool:
                    """Returns whether the model can optionally use node types."""
                    return None

                @classmethod
                def requires_edge_types(cls) -> str:
                    """Returns whether the model requires edge types."""
                    return None
                
                @classmethod
                def can_use_edge_types(cls) -> bool:
                    """Returns whether the model can optionally use edge types."""
                    return None

                @classmethod
                def requires_positive_edge_weights(cls) -> str:
                    """Returns whether the model requires positive edge weights."""
                    return None

                @classmethod
                def requires_edge_weights(cls) -> str:
                    """Returns whether the model can optionally use edge weights."""
                    return None

                @classmethod
                def can_use_edge_weights(cls) -> bool:
                    """Returns whether the model can optionally use edge weights."""
                    return None

                @classmethod
                def is_stocastic(cls) -> bool:
                    """Returns whether the model can optionally use edge weights."""
                    return None
                
                @classmethod
                def requires_edge_type_features(cls) -> str:
                    """Returns whether the model requires edge type features."""
                    return None
                
                @classmethod
                def can_use_edge_type_features(cls) -> bool:
                    """Returns whether the model can optionally use edge type features."""
                    return None
                
                @classmethod
                def requires_edge_features(cls) -> str:
                    """Returns whether the model requires edge features."""
                    return None
                
                @classmethod
                def can_use_edge_features(cls) -> bool:
                    """Returns whether the model can optionally use edge features."""
                    return None

                @classmethod
                def is_topological(cls) -> str:
                    """Returns whether this embedding is based on graph topology."""
                    return parent_class.is_topological()

                @classmethod
                def task_name(cls) -> str:
                    """Returns name of the task"""
                    return task_name

                @classmethod
                def smoke_test_parameters(cls) -> Dict[str, Any]:
                    """Returns parameters for the smoke test."""
                    return dict()

                @classmethod
                def is_available(cls) -> bool:
                    """Returns whether the model class is actually available in the user system."""
                    return False

            model_class = StubClass
        else:
            # We re-raise the exception.
            raise exception

    # We assign the local class as exposed in the
    # caller frame.
    frame.f_back.f_locals[model_class_name] = model_class

    # We register the newly loaded class.
    model_class.register(model_class)



"""Unit test class for testing corner cases in edge-label GCN model."""
import os
from unittest import TestCase

import compress_json
import numpy as np
from dict_hash import sha256
from ensmallen import Graph
from tqdm import tqdm

from embiggen.edge_prediction import GCNEdgePrediction
from embiggen.embedders import HyperSketching
from embiggen.embedders.ensmallen_embedders.hyper_sketching import \
    HyperSketching
from .cached_tests import cache_or_store


class TestGCNEdgePrediction(TestCase):
    """Unit test class for edge-label prediction pipeline."""

    def setUp(self):
        """Setup objects for running tests on edge-label prediction pipeline class."""
        graph = Graph.generate_random_connected_graph(
            node_type="hallo", edge_type="huhu", weight=56.0, number_of_nodes=20
        ) | Graph.generate_random_connected_graph(
            node_type="hallo2", edge_type="huhu2", weight=56.0, number_of_nodes=20
        )
        self.graph = graph.remove_parallel_edges().add_selfloops("huhu2", 3.0)

    def test_evaluate_embedding_for_edge_label_prediction(self):
        """Test graph visualization."""
        if not GCNEdgePrediction.is_available():
            return

        if cache_or_store([
            "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
            "./embiggen/edge_prediction/edge_prediction_tensorflow/gcn.py",
            "./embiggen/utils/abstract_gcn.py",
            "./embiggen/utils/abstract_edge_gcn.py",
        ]):
            return

        parametrizations = []
        features = []

        edge_features = HyperSketching()
        edge_features.fit(self.graph)

        rnd = np.random.RandomState(42)

        for use_node_embedding in (True, False):
            for use_node_type_embedding in (True, False):
                for residual_convolutional_layers in (True, False):
                    for kernels in (
                        "Left Normalized Laplacian",
                        ["Left Normalized Laplacian", "Right Normalized Laplacian"],
                        None,
                    ):
                        for use_edge_type_embedding in (True, False):
                            for use_edge_metrics in (True, False):
                                for node_type_features in [
                                    rnd.random(
                                        (self.graph.get_number_of_node_types(), 5)
                                    ),
                                    [
                                        rnd.random(
                                            (self.graph.get_number_of_node_types(), 3)
                                        ),
                                        rnd.random(
                                            (self.graph.get_number_of_node_types(), 7)
                                        ),
                                    ],
                                    None,
                                ]:
                                    for node_features in [
                                        rnd.random(
                                            (self.graph.get_number_of_nodes(), 5)
                                        ),
                                        [
                                            rnd.random(
                                                (self.graph.get_number_of_nodes(), 13)
                                            ),
                                            rnd.random(
                                                (self.graph.get_number_of_nodes(), 17)
                                            ),
                                        ],
                                        None,
                                    ]:
                                        for edge_type_features in [
                                            rnd.random(
                                                (
                                                    self.graph.get_number_of_edge_types(),
                                                    15,
                                                )
                                            ),
                                            [
                                                rnd.random(
                                                    (
                                                        self.graph.get_number_of_edge_types(),
                                                        9,
                                                    )
                                                ),
                                                rnd.random(
                                                    (
                                                        self.graph.get_number_of_edge_types(),
                                                        11,
                                                    )
                                                ),
                                            ],
                                            None,
                                        ]:
                                            for edge_features in [edge_features, None]:
                                                for siamese_node_feature_module in (True, False):
                                                    if (
                                                        not use_node_embedding
                                                        and not use_node_type_embedding
                                                        and not use_edge_type_embedding
                                                        and not use_edge_metrics
                                                        and node_type_features is None
                                                        and node_features is None
                                                        and edge_features is None
                                                        and edge_type_features is None
                                                    ):
                                                        continue
                                                    features.append(dict(
                                                        node_type_features=node_type_features,
                                                        node_features=node_features,
                                                        edge_type_features=edge_type_features,
                                                        edge_features=edge_features,
                                                    ))
                                                    parametrizations.append(dict(
                                                        use_node_embedding=use_node_embedding,
                                                        use_node_type_embedding=use_node_type_embedding,
                                                        residual_convolutional_layers=residual_convolutional_layers,
                                                        siamese_node_feature_module=siamese_node_feature_module,
                                                        kernels=kernels,
                                                        use_edge_type_embedding=use_edge_type_embedding,
                                                        use_edge_metrics=use_edge_metrics,
                                                    ))
        for parametrization, feature_set in tqdm(
            zip(parametrizations, features),
            total=len(parametrizations),
            desc="Testing parametrization combo",
        ):
            if not parametrization["use_node_embedding"] and not parametrization["use_node_type_embedding"] and feature_set["node_type_features"] is None and feature_set["node_features"] is None:
                parametrization["kernels"] = None
                parametrization["edge_embedding_methods"] = None

            if parametrization["kernels"] is None:
                parametrization["residual_convolutional_layers"] = False
            
            salt = sha256(
                dict(
                    parametrization=parametrization,
                    feature_set=feature_set
                ),
                use_approximation=True
            )

            if cache_or_store(
                [
                    "./embiggen/embedders/ensmallen_embedders/hyper_sketching.py",
                    "./embiggen/edge_prediction/edge_prediction_tensorflow/gcn.py",
                    "./embiggen/utils/abstract_gcn.py",
                    "./embiggen/utils/abstract_edge_gcn.py",
                ],
                salt=salt
            ):
                continue

            model: GCNEdgePrediction = GCNEdgePrediction(
                epochs=1,
                number_of_batches_per_epoch=1,
                number_of_units_per_graph_convolution_layers=4,
                number_of_units_per_ffnn_body_layer=4,
                number_of_units_per_ffnn_head_layer=4,
                number_of_graph_convolution_layers=2
                if parametrization["kernels"] is not None
                else 0,
                node_embedding_size=5,
                node_type_embedding_size=5,
                edge_type_embedding_size=5,
                **parametrization
            )
            model.fit(graph=self.graph, **feature_set)
            model.predict(graph=self.graph, **feature_set)
            model.predict_proba(graph=self.graph, **feature_set)

            compress_json.dump(
                parametrization,
                path,
            )


"""Module providing HyperSketching implementation."""
from typing import Dict, Any, Optional, Tuple, List
from ensmallen import Graph
import pandas as pd
import numpy as np
import compress_json
import json
from ensmallen import models
from embiggen.embedders.ensmallen_embedders.ensmallen_embedder import EnsmallenEmbedder
from embiggen.utils import EmbeddingResult
from embiggen.utils import AbstractEdgeFeature


class HyperSketching(EnsmallenEmbedder, AbstractEdgeFeature):
    """Class implementing the HyperSketching edge embedding method."""

    def __init__(
        self,
        number_of_hops: int = 3,
        precision: int = 6,
        bits: int = 4,
        include_node_types: bool = False,
        include_edge_types: bool = False,
        include_edge_ids: bool = False,
        include_node_ids: bool = True,
        include_selfloops: bool = True,
        include_typed_graphlets: bool = False,
        random_state: int = 42,
        number_of_random_integers: int = 0,
        normalize_by_symmetric_laplacian: bool = True,
        concatenate_features: bool = False,
        dtype: str = "f32",
        overlap_path: Optional[str] = None,
        left_difference_path: Optional[str] = None,
        right_difference_path: Optional[str] = None,
        ring_bell: bool = False,
        enable_cache: bool = False
    ):
        """Create new HyperSketching model.

        Parameters
        --------------------------
        number_of_hops: int = 2
            The number of hops for the Sketches.
        precision: int = 6
            The precision of the HyperLogLog counters.
            The supported values range from 4 to 16.
        bits: int = 5
            The number of bits of the HyperLogLog counters.
            The supported values range from 4 to 6.
        include_node_types: bool = False,
            Whether to include node types in the sketches.
        include_edge_types: bool = False,
            Whether to include edge types in the sketches.
        include_edge_ids: bool = False,
            Whether to include edge ids in the sketches.
        include_node_ids: bool = True,
            Whether to include node ids in the sketches.
        include_selfloops: bool = True,
            Whether to include selfloops in the sketches.
        include_typed_graphlets: bool = False,
            Whether to include typed graphlets in the sketches.
        random_state: int = 42,
            The random state to use.
        number_of_random_integers: int = 0,
            The number of random integers to use per node.
        normalize_by_symmetric_laplacian: bool = True,
            Whether to normalize the sketches by the symmetric laplacian.
        concatenate_features: bool = False,
            Whether to Concatenate the normalized and non-normalized features.
        dtype: str = "f32",
            The type of the features.
        overlap_path: Optional[str] = None,
            The path to the overlap file.
            This will be the position where, if provided, we will MMAP
            the overlap numpy array.
        left_difference_path: Optional[str] = None,
            The path to the left difference file.
            This will be the position where, if provided, we will MMAP
            the left difference numpy array.
        right_difference_path: Optional[str] = None,
            The path to the right difference file.
            This will be the position where, if provided, we will MMAP
            the right difference numpy array.
        ring_bell: bool = False,
            Whether to ring the bell when the sketches are ready.
        enable_cache: bool = False,
            Whether to enable caching of the sketches.
        """
        self._kwargs = dict(
            number_of_hops=number_of_hops,
            precision=precision,
            bits=bits,
            include_node_types=include_node_types,
            include_edge_types=include_edge_types,
            include_edge_ids=include_edge_ids,
            include_node_ids=include_node_ids,
            include_selfloops=include_selfloops,
            include_typed_graphlets=include_typed_graphlets,
            number_of_random_integers=number_of_random_integers,
            normalize_by_symmetric_laplacian=normalize_by_symmetric_laplacian,
            concatenate_features=concatenate_features,
            dtype=dtype,
        )

        self._overlap_path=overlap_path
        self._left_difference_path=left_difference_path
        self._right_difference_path=right_difference_path

        self._model = models.HyperSketching(
            **self._kwargs,
            random_state=random_state,
        )

        self._fitting_was_executed = False

        super().__init__(
            enable_cache=enable_cache,
            ring_bell=ring_bell,
            random_state=random_state,
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters of the model."""
        return dict(
            **super().parameters(),
            **self._kwargs,
            overlap_path=self._overlap_path,
            left_difference_path=self._left_difference_path,
            right_difference_path=self._right_difference_path,
        )

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(
            number_of_hops=2,
            precision=4,
            bits=4,
        )
    
    def is_fit(self) -> bool:
        """Return whether the model was fit."""
        return self._fitting_was_executed
    
    def fit(
        self,
        graph: Graph,
    ):
        """Fit the model on the provided graph.

        Parameters
        -------------------
        graph: Graph,
            The graph to fit the model on.
        """
        self._fitting_was_executed = True
        self._model.fit(graph)
        return self
        
    def get_bits(self):
        """Return the number of bits used for the HyperLogLog counters."""
        return self._model.get_bits()
    
    def get_precision(self):
        """Return the precision used for the HyperLogLog counters."""
        return self._model.get_precision()
        
    def get_number_of_hops(self):
        """Return the number of hops used for the sketches."""
        return self._model.get_number_of_hops()
    
    @classmethod
    def get_feature_dictionary_keys(cls) -> List[str]:
        """Return the list of keys to be used in the feature dictionary."""
        return [
            "overlap",
            "left_difference",
            "right_difference",
        ]
    
    def get_feature_dictionary_shapes(self) -> Dict[str, List[int]]:
        """Return the dictionary of shapes to be used in the feature dictionary."""
        factor = 2 if self.get_concatenate_features() else 1
        return dict(
            overlap=[factor * self.get_number_of_hops()**2],
            left_difference=[factor * self.get_number_of_hops()],
            right_difference=[factor * self.get_number_of_hops()],
        )
    
    def get_difference_cardinalities_from_node_ids(self, src: int, dst: int)-> np.ndarray:
        """Return the cardinalities of the differences between the sketches of the two nodes.
        
        Parameters
        -------------------
        src: int,
            The source node id.
        dst: int,
            The destination node id.

        Returns
        -------------------
        The cardinalities of the differences between the sketches of the two nodes.
        This is a numpy array of shape (self.get_number_of_hops(), ).

        Raises
        -------------------
        ValueError,
            If the provided node ids are not in the graph.
            If the model was not fitted.
        """
        return self._model.get_difference_cardinalities_from_node_ids(src, dst)
    
    def get_overlap_cardinalities_from_node_ids(self, src: int, dst: int) -> np.ndarray:
        """Return the cardinalities of the overlaps between the sketches of the two nodes.

        Parameters
        -------------------
        src: int,
            The source node id.
        dst: int,
            The destination node id.

        Returns
        -------------------
        The cardinalities of the overlaps between the sketches of the two nodes.
        This is a numpy array of shape (self.get_number_of_hops(), self.get_number_of_hops()).

        Raises
        -------------------
        ValueError,
            If the provided node ids are not in the graph.
            If the model was not fitted.
        """
        return self._model.get_overlap_cardinalities_from_node_ids(src, dst)

    def get_normalize_by_symmetric_laplacian(self) -> bool:
        """Return whether the sketches are normalized by the symmetric laplacian."""
        return self._model.get_normalize_by_symmetric_laplacian()
    
    def get_concatenate_features(self) -> bool:
        """Return whether the features are Concatenated."""
        return self._model.get_concatenate_features()

    def _fit_transform(
        self,
        graph: Graph,
        return_dataframe: bool = True,
    ) -> EmbeddingResult:
        """Return edge sketches.
        
        Parameters
        -------------------
        graph: Graph,
            The graph to fit the model on.
        return_dataframe: bool = True,
            Whether to return the results as pandas dataframes.
        """
        if not self._fitting_was_executed:
            self.fit(graph)
        (overlaps, left_difference, right_difference) = self._model.get_sketching_for_all_edges(
            graph,
            support=graph,
            overlap_path=self._overlap_path,
            left_difference_path=self._left_difference_path,
            right_difference_path=self._right_difference_path,
        )
        if return_dataframe:
            factor = 2 if self.get_concatenate_features() else 1
            extra_columns = ["N"] if self.get_concatenate_features() else []
            overlaps = pd.DataFrame(
                overlaps.reshape(-1, factor * self.get_number_of_hops()**2),
                index=graph.get_edge_node_names(),
                columns=[
                    f"{column}_{i}{j}"
                    for column in ["A"] + extra_columns
                    for i in range(self.get_number_of_hops())
                    for j in range(self.get_number_of_hops())
                ]
            )
            left_difference = pd.DataFrame(
                left_difference.reshape(-1, factor * self.get_number_of_hops()),
                index=graph.get_edge_node_names(),
                columns=[
                    f"{column}_{i}"
                    for column in ["L"] + extra_columns
                    for i in range(self.get_number_of_hops())
                ]
            )
            right_difference = pd.DataFrame(
                right_difference.reshape(-1, factor * self.get_number_of_hops()),
                index=graph.get_edge_node_names(),
                columns=[
                    f"{column}_{i}"
                    for column in ["R"] + extra_columns
                    for i in range(self.get_number_of_hops())
                ]
            )

        return EmbeddingResult(
            embedding_method_name=self.model_name(),
            edge_embeddings=[overlaps, left_difference, right_difference],
        )
    
    def get_sketching_from_edge_node_ids(
        self,
        support: Graph,
        sources: np.ndarray,
        destinations: np.ndarray,
        overlap_path: Optional[str] = None,
        left_difference_path: Optional[str] = None,
        right_difference_path: Optional[str] = None,
    ) -> Tuple[np.ndarray]:
        """Return the sketches for the provided edges.
        
        Parameters
        -------------------
        support: Graph,
            The graph from which we extract the node degrees if the
            laplacian normalization is enabled. Be advised that this
            graph should, in most cases, be the same as the one used
            to fit the model.
        sources: np.ndarray,
            The source node ids.
        destinations: np.ndarray,
            The destination node ids.
        overlap_path: Optional[str] = None,
            The path to the overlap file.
            If an overlap path was provided in the constructor and this
            parameter is None, then the overlap will be loaded from the
            file provided in the constructor.
            This will be the position where, if provided, we will MMAP
            the overlap numpy array.
        left_difference_path: Optional[str] = None,
            The path to the left difference file.
            If a left difference path was provided in the constructor and this
            parameter is None, then the left difference will be loaded from the
            file provided in the constructor.
            This will be the position where, if provided, we will MMAP
            the left difference numpy array.
        right_difference_path: Optional[str] = None,
            The path to the right difference file.
            If a right difference path was provided in the constructor and this
            parameter is None, then the right difference will be loaded from the
            file provided in the constructor.
            This will be the position where, if provided, we will MMAP
            the right difference numpy array.

        Returns
        -------------------
        The sketches for the provided edges.

        Raises
        -------------------
        ValueError,
            If the provided node ids are not in the graph.
            If the model was not fitted.
        """
        if not self._fitting_was_executed:
            raise ValueError(
                "The model was not fitted."
            )

        if overlap_path is None:
            overlap_path = self._overlap_path

        if left_difference_path is None:
            left_difference_path = self._left_difference_path
        
        if right_difference_path is None:
            right_difference_path = self._right_difference_path

        return self._model.get_sketching_from_edge_node_ids(
            support,
            sources,
            destinations,
            overlap_path=overlap_path,
            left_difference_path=left_difference_path,
            right_difference_path=right_difference_path,
        )
    
    def get_edge_feature_from_edge_node_ids(
        self,
        support: Graph,
        sources: np.ndarray,
        destinations: np.ndarray,
    ) -> Dict[str, np.ndarray]:
        """Return the edge feature for the given edge.
        
        Parameters
        -----------------------
        support: Graph,
            The graph to use as base for the topological metrics.
        sources: np.ndarray,
            The source node ids.
        destinations: np.ndarray,
            The destination node ids.
        """
        overlap, left_difference, right_difference = self.get_sketching_from_edge_node_ids(
            support,
            sources,
            destinations,
        )
        return dict(
            overlap=overlap,
            left_difference=left_difference,
            right_difference=right_difference,
        )
    
    def get_edge_feature_from_graph(self, graph: Graph, support: Graph) -> Dict[str, np.ndarray]:
        """Return the edge feature for the given graph.
        
        Parameters
        -----------------------
        graph: Graph,
            The graph to use as base for the topological metrics.
        support: Graph,
            The graph to use as base for the topological metrics.
        """
        if not self._fitting_was_executed:
            raise ValueError(
                "The model was not fitted."
            )
        overlap, left_difference, right_difference = self._model.get_sketching_for_all_edges(
            graph,
            support=support,
            overlap_path=self._overlap_path,
            left_difference_path=self._left_difference_path,
            right_difference_path=self._right_difference_path,
        )

        # A small debug assert to ensure the APIs are not broken.
        for feature in (overlap, left_difference, right_difference):
            assert feature.shape[0] == graph.get_number_of_edges()

        return dict(
            overlap=overlap,
            left_difference=left_difference,
            right_difference=right_difference,
        )

    @classmethod
    def get_feature_name(cls) -> str:
        """Return the feature names."""
        return cls.model_name()
    
    @classmethod
    def task_name(cls) -> str:
        return "Edge Embedding"

    @classmethod
    def model_name(cls) -> str:
        """Returns name of the model."""
        return "HyperSketching"
    
    @classmethod
    def can_use_edge_weights(cls) -> bool:
        """Returns whether the model can optionally use edge weights."""
        return False

    @classmethod
    def can_use_node_types(cls) -> bool:
        """Returns whether the model can optionally use node types."""
        return True

    @classmethod
    def requires_node_types(cls) -> bool:
        """Returns whether the model requires node types."""
        return False
    
    def is_using_node_types(self) -> bool:
        """Returns whether the model is using node types."""
        return self._kwargs["include_node_types"]

    @classmethod
    def can_use_edge_types(cls) -> bool:
        """Returns whether the model can optionally use edge types."""
        return True

    @classmethod
    def requires_edge_types(cls) -> bool:
        """Returns whether the model requires edge types."""
        return False
    
    def is_using_edge_types(self) -> bool:
        """Returns whether the model is using edge types."""
        return self._kwargs["include_edge_types"]
    
    @classmethod
    def is_stocastic(cls) -> bool:
        """Returns whether the model is stocastic and has therefore a random state."""
        return True
    
    def clone(self) -> "Self":
        """Return a fresh clone of the model."""
        return HyperSketching(**self.parameters())
    
    @classmethod
    def load(cls, path: str) -> "Self":
        """Load a saved version of the model from the provided path.

        Parameters
        -------------------
        path: str
            Path from where to load the model.
        """
        data = compress_json.load(path)
        model = HyperSketching(**data["parameters"])
        model._model = models.HyperSketching.loads(
            json.dumps(data["inner_model"])
        )
        for key, value in data["metadata"].items():
            model.__setattr__(key, value)
        return model

    def dumps(self) -> Dict[str, Any]:
        """Dumps the current model as dictionary."""
        return dict(
            parameters=self.parameters(),
            inner_model=json.loads(self._model.dumps()),
            metadata=dict(
                _fitting_was_executed=self._fitting_was_executed
            )
        )

    def dump(self, path: str):
        """Dump the current model at the provided path.

        Parameters
        -------------------
        path: str
            Path from where to dump the model.
        """
        compress_json.dump(
            self.dumps(),
            path
        )



"""Submodule wrapping Linear Support Vector Machine for Edge prediction."""
from typing import Dict, Any, Union, List
from sklearn.svm import LinearSVC
from embiggen.edge_prediction.edge_prediction_sklearn.sklearn_edge_prediction_adapter import SklearnEdgePredictionAdapter
from embiggen.utils import normalize_kwargs


class LinearSVCEdgePrediction(SklearnEdgePredictionAdapter):
    """Create wrapper over Sklearn Linear Support Vector Machine classifier for Edge prediction."""

    def __init__(
        self,
        penalty: str = "l2",
        loss: str = "squared_hinge",
        dual: bool = True,
        tol: float = 1e-4,
        C: float = 1.0,
        multi_class: str = "ovr",
        fit_intercept: bool = True,
        intercept_scaling: float = 1.0,
        class_weight: Union[Dict, str] = "balanced",
        verbose: int = 0,
        max_iter: int = 1000,
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        training_unbalance_rate: float = 1.0,
        training_sample_only_edges_with_heterogeneous_node_types: bool = False,
        use_edge_metrics: bool = False,
        use_scale_free_distribution: bool = True,
        prediction_batch_size: int = 2**12,
        random_state: int = 42
    ):
        """Create the Linear Support Vector Machine for Node-label prediction."""
        self._kwargs = normalize_kwargs(
            self,
            dict(
                penalty=penalty,
                loss=loss,
                dual=dual,
                tol=tol,
                C=C,
                multi_class=multi_class,
                fit_intercept=fit_intercept,
                intercept_scaling=intercept_scaling,
                class_weight=class_weight,
                verbose=verbose,
                max_iter=max_iter,
            )
        )

        super().__init__(
            LinearSVC(**self._kwargs, random_state=random_state),
            edge_embedding_methods=edge_embedding_methods,
            training_unbalance_rate=training_unbalance_rate,
            use_edge_metrics=use_edge_metrics,
            use_scale_free_distribution=use_scale_free_distribution,
            training_sample_only_edges_with_heterogeneous_node_types=training_sample_only_edges_with_heterogeneous_node_types,
            prediction_batch_size=prediction_batch_size,
            random_state=random_state
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        return {**super().parameters(), **self._kwargs}

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(max_iter=2)

    @classmethod
    def model_name(cls) -> str:
        return "Linear Support Vector Classifier"



from typing import Dict, Any
import pytest
from embiggen.embedders import embed_graph
from embiggen.utils import AbstractEmbeddingModel
from embiggen.embedders.ensmallen_embedders.hope import HOPEEnsmallen
from unittest import TestCase


class ClassForTestAbstractEmbeddingModel(AbstractEmbeddingModel):

    def __init__(self):
        super().__init__(
            embedding_size=100
        )

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(
            invalid_parameter=5,
        )

    @classmethod
    def is_stocastic(cls) -> bool:
        """Returns whether the model is stocastic and has therefore a random state."""
        return False

    @classmethod
    def model_name(cls) -> str:
        """Returns name of the model."""
        return "TMP"

    @classmethod
    def library_name(cls) -> str:
        """Returns name of the library."""
        return "TMP"

    @classmethod 
    def can_use_edge_weights(cls) -> bool:
        return False

    @classmethod 
    def can_use_node_types(cls) -> bool:
        return False

    @classmethod 
    def can_use_edge_types(cls) -> bool:
        return False
    
    @classmethod
    def requires_nodes_sorted_by_decreasing_node_degree(cls) -> bool:
        return False

class TestEmbedGraph(TestCase):

    def setUp(self):
        pass

    def test_embed_graph(self):
        with pytest.raises(ValueError):
            embed_graph(
                "Cora",
                embedding_model=int,
                repository="linqs"
            )

        with pytest.raises(ValueError):
            embed_graph(
                "Cora",
                embedding_model=HOPEEnsmallen(),
                repository="linqs",
                embedding_size=1000
            )
        
        with pytest.raises(ValueError):
            embed_graph(
                "Cora",
                embedding_model=ClassForTestAbstractEmbeddingModel(),
                repository="linqs",
                smoke_test=True
            )

        with pytest.raises(ValueError):
            embed_graph(
                "Cora",
                embedding_model=ClassForTestAbstractEmbeddingModel(),
                repository="linqs"
            )

        embed_graph(
            "Cora",
            embedding_model="Degree-based SPINE",
            repository="linqs"
        )

        


"""Submodule wrapping Nu Support Vector Machine for Edge-label prediction."""
from typing import Dict, Any, Union, List
from sklearn.svm import NuSVC
from embiggen.node_label_prediction.node_label_prediction_sklearn.sklearn_node_label_prediction_adapter import (
    SklearnNodeLabelPredictionAdapter,
)
from embiggen.utils import normalize_kwargs


class NuSVCEdgeLabelPrediction(SklearnNodeLabelPredictionAdapter):
    """Create wrapper over Sklearn Nu Support Vector Machine classifier for Edge-label prediction."""

    def __init__(
        self,
        nu: float = 0.1,
        kernel: str = "rbf",
        degree: int = 3,
        gamma: str = "scale",
        coef0: float = 0.0,
        shrinking: bool = True,
        probability: bool = False,
        tol: float = 1e-3,
        cache_size: int = 200,
        class_weight: Union[Dict, str] = "balanced",
        verbose: bool = False,
        max_iter: int = -1,
        decision_function_shape: str = "ovr",
        break_ties: bool = False,
        edge_embedding_methods: Union[List[str], str] = "Concatenate",
        use_edge_metrics: bool = False,
        random_state: int = 42,
    ):
        """Create the Nu SVC for Edge-label prediction."""
        self._kwargs = normalize_kwargs(
            self,
            dict(
                nu=nu,
                kernel=kernel,
                degree=degree,
                gamma=gamma,
                coef0=coef0,
                shrinking=shrinking,
                probability=probability,
                tol=tol,
                cache_size=cache_size,
                class_weight=class_weight,
                verbose=verbose,
                max_iter=max_iter,
                decision_function_shape=decision_function_shape,
                break_ties=break_ties,
            )
        )

        super().__init__(
            NuSVC(**self._kwargs, random_state=random_state),
            edge_embedding_methods=edge_embedding_methods,
            use_edge_metrics=use_edge_metrics,
            random_state=random_state,
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        return {**super().parameters(), **self._kwargs}
    
    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(max_iter=1)

    @classmethod
    def model_name(cls) -> str:
        return "Nu Support Vector Classifier"



"""Submodule providing edge-label prediction models based on LightGBM Models."""
from embiggen.edge_label_prediction.sklearn_like_edge_label_prediction_adapter import (
    SklearnLikeEdgeLabelPredictionAdapter,
)
from embiggen.utils.abstract_models import build_init

build_init(
    module_library_names="lightgbm",
    formatted_library_name="LightGBM",
    task_name="Edge Label Prediction",
    expected_parent_class=SklearnLikeEdgeLabelPredictionAdapter,
)



"""Submodule wrapping Ridge Classifier for Node-label prediction."""
from typing import Dict, Any, Union, Optional
from sklearn.linear_model import RidgeClassifier
from embiggen.node_label_prediction.node_label_prediction_sklearn.sklearn_node_label_prediction_adapter import (
    SklearnNodeLabelPredictionAdapter,
)
from embiggen.utils import normalize_kwargs


class RidgeClassifierNodeLabelPrediction(SklearnNodeLabelPredictionAdapter):
    """Create wrapper over Sklearn Ridge Classifier classifier for Node-label prediction."""

    def __init__(
        self,
        alpha: float = 1.0,
        fit_intercept: bool = True,
        copy_X: bool = True,
        max_iter: Optional[int] = None,
        tol: float = 1e-4,
        class_weight: Union[Dict, str] = "balanced",
        solver: str = "auto",
        positive: bool = False,
        random_state: int = 323,
    ):
        """Create the Ridge Classifier for Node-label prediction."""
        self._kwargs = normalize_kwargs(
            self,
            dict(
                alpha=alpha,
                fit_intercept=fit_intercept,
                copy_X=copy_X,
                max_iter=max_iter,
                tol=tol,
                class_weight=class_weight,
                solver=solver,
                positive=positive,
            )
        )

        super().__init__(
            RidgeClassifier(**self._kwargs, random_state=random_state),
            random_state=random_state,
        )

    def parameters(self) -> Dict[str, Any]:
        """Returns parameters used for this model."""
        return {**super().parameters(), **self._kwargs}

    @classmethod
    def smoke_test_parameters(cls) -> Dict[str, Any]:
        """Returns parameters for smoke test."""
        return dict(max_iter=2)

    @classmethod
    def model_name(cls) -> str:
        return "Ridge Classifier"
